/***************************************************************************
 *   Copyright (C) 2005 by Peter Müller   *
 *   pmueller@cs.tu-berlin.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "comet.h"

PComet::PComet(CometenScene * _scene, float _time, float _life_time, int _num_children,
               const Vec3 & _p, const Vec3 & _v, const Vec3 & _a, float _size, 
               const Vec4 & _acolor, const Vec4 & _ecolor):
Comet(_scene, _p, _v),
accel(_a),
size(_size),
num_children(_num_children),
acolor(_acolor),
ecolor(_ecolor),
start_time(_time),
life_time(_life_time),
psystem(NULL),
decal(NULL),
_destroyed(false)
{
	GLuint tex = scene->getTexture("particle.png");
	radius = size;
	
	psystem = 
			new ParticleSystem(tex, pos, 300, 0, 0.5f, 1.0f,  true, 350, Vec3::ZERO, 
							8*size, 16*size, false, 4*size, acolor, ecolor, _time);
	
	decal = new Decal(tex, pos, true, Vec3::X, Vec2(10*size,10*size), acolor, _time, 0, 0.5f);
}

PComet::~PComet() {
	delete psystem;
	delete decal;
}

void PComet::process(float time, float deltat) {
	decal->setPos(pos);
	decal->process(time);
	psystem->setPos(pos);
	psystem->process(time);
	
	vel += accel * deltat;

	if (time - start_time > life_time && !destroyed()) 
		destroy(time);
}

void PComet::render() {
	decal->render(scene->getCamPos());
	psystem->render(scene->getCamPos());
}

void PComet::hitPlane(float time, const Plane & plane, const Vec3 & hit_point) {	
	// bounce comet
	vel.reflect(plane.N);
	
	float decal_size = 1.0f * fabs(vel.dot(plane.N)) * radius;
	Effect * decal = new TimeLimitDecal(scene->getTexture("particle.png"), hit_point, false, plane.N, 
										Vec2(decal_size/2,decal_size/2), 
										Vec2(decal_size,decal_size), acolor, time, 1.5f, 0.125f, 1.375f);
	scene->addEffect(decal);
}

void PComet::destroy(float time) {
	Comet * newc = NULL;
	Effect * newe = NULL;
	
	if (destroyed())
		return;
	
	for (int i = 0; i < num_children; i++) {
		newc = new PComet(scene, time, frand(0.5f, 1.0f), 0, pos,
						  vel / 3 + Vec3::randomUnit() * frand(30.0f, 60.0f), 
		                  Vec3::ZERO, size / 3, acolor, ecolor);
		scene->addComet(newc);
	}
	
	newe = new Explosion(scene->getTexture("particle.png"), scene->getTexture("particle.png"),
						acolor, ecolor,
						pos, Vec3::randomUnit(), 25*size, true, time);
	
	scene->addEffect(newe);

	kill();
}

void PComet::kill() {
	psystem->stop();
	decal->stop();
	_destroyed = true;
}

bool PComet::destroyed() {
	return _destroyed;
}

bool PComet::done() {
	return _destroyed && psystem->done() && decal->done();
}

