/***************************************************************************
 *   Copyright (C) 2005 by Peter Müller   *
 *   pmueller@cs.tu-berlin.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "cometenscene.h"
#include "comet.h"

// ============================= init/ quit ================================

void CometenScene::init(float time, GLuint bg, int _colors, float _create_intervall, 
                        float _lightmap_alpha, bool _do_matrix, bool _do_slowmotion, bool _do_blitz) 
{
	quit();
    glShadeModel(GL_SMOOTH);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClearDepth(1.0f);
    glDepthFunc(GL_LEQUAL);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
   	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
   	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
   	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glFrontFace(GL_CW);
	
	glPointParameterf(GL_POINT_FADE_THRESHOLD_SIZE, 60.0f);
	glPointParameterf(GL_POINT_SIZE_MIN, 0.0f);
	glPointParameterf(GL_POINT_SIZE_MAX, 1000.0f);

	box_height = box_size;
	box_cam_distance = 0.0f;

	loadTexture("particle.png");
	loadTexture("lightmap.png");
	
	lightmap_alpha = _lightmap_alpha;
	background = bg;
	lightmap = getTexture("lightmap.png");
	create_intervall = _create_intervall;
	colors = _colors;
	
	do_matrix = _do_matrix;
	do_slowmotion = _do_slowmotion;
	do_blitz = _do_blitz;
	
	scenetimer.setScale(1.0f, time);
	scenetimer.setTime(0.0f, time);
	
	last_comet_process = scenetimer.getTime(time);
	next_comet = scenetimer.getTime(time) + 1.25f;
	next_blitz_effect = scenetimer.getTime(time) + frand(15.0f, 45.0f);
	
	next_cam_effect = time + frand(60.0f, 120.0f);
	
	state = S_NORMAL;
}

void CometenScene::resize(int res_x, int res_y) {
	// a, b, c: size' = size * sqrt[1 / (a + b*d + c*d^2)]
	float dist_attenuation[] = {
		0.0f, 
		0.0f, 
		4.0f * tanf(deg2rad(fov/2)) * tanf(deg2rad(fov/2)) /(res_y*res_y)
	};
	
	glPointParameterfv(GL_POINT_DISTANCE_ATTENUATION, dist_attenuation);
		
   	glViewport(0, 0, res_x, res_y);	
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov, (float)res_x / res_y, 1.0f, 1000.0f);
	glMatrixMode(GL_MODELVIEW);
	
	box_height = box_size / ((float)res_x / res_y);
	box_cam_distance = box_height / 2 / tanf(deg2rad(fov/2));
	
	planes.clear();
	
	planes.push_back(Plane(Vec3::X, -box_size/2));
	planes.push_back(Plane(-Vec3::X, -box_size/2));
	planes.push_back(Plane(Vec3::Y, -box_height/2));
	planes.push_back(Plane(-Vec3::Y, -box_height/2));
	planes.push_back(Plane(Vec3::Z, -box_size/2));
	planes.push_back(Plane(-Vec3::Z, -box_size/2));	
}

void CometenScene::quit() {
	list<Comet*>::iterator j;
	list<Effect*>::iterator i;
	
	freeTextures();
	
	for (j = comets.begin(); j != comets.end(); j++) {
		delete (*j); 
		*j = NULL;
	}
	
	for (i = effects.begin(); i != effects.end(); i++) {
		delete (*i); 
		*i = NULL;
	}
	
	planes.clear();
}

// ========================== process ==============================

void CometenScene::process(float time) {
	switch (state) {
	case S_NORMAL:
		processNormal(time);
		break;
	case S_MATRIX:
		processMatrix(time);
		break;
	case S_SLOWMOTION:
		processSlowmotion(time);
		break;
	}
	
	processScene(scenetimer.getTime(time));
	
	if (state == S_NORMAL && time > next_cam_effect) {
		if (do_matrix && do_slowmotion) {
			switch(rand() % 2) {
			case 0:
				state = S_SLOWMOTION;
				state_start = time;
				break;
			case 1:
			default:
				state = S_MATRIX;
				state_start = time;
				break;
			}	
		}
		else if (do_matrix) {
			state = S_MATRIX;
			state_start = time;
		}
		else if (do_slowmotion) {
			state = S_SLOWMOTION;
			state_start = time;
		}

		next_cam_effect = time + frand(120.0f, 180.0f);
	}	
}

void CometenScene::processNormal(float time) {
	cam_viewpoint = Vec3::ZERO;
	campos = Vec3(0,0,box_size/2 + box_cam_distance);
	fade = 0.0f;
	black_bars = false;
}

void CometenScene::processMatrix(float time) {
	float t = time - state_start;
	
	cam_viewpoint = Vec3::ZERO;
	
	if (t < 0.5f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = (t) / 0.5f;
		black_bars = false;
		scenetimer.setScale(1.0f, time);
	}
	else if (t < 1.0f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = 1.0f;
		black_bars = true;
		scenetimer.setScale(0.1f, time);
	}
	else if (t < 31.0f) {
		black_bars = true;
		scenetimer.setScale(0.1f, time);
		if (t < 1.5f) {
			fade = 1 - (t - 1.0f) / 0.5f;
		}
		else if (t > 30.5f) {
			fade = (t - 30.5f) / 0.5f;
		}
		else {
			fade = 0.0f;
		}
		
		float w = (t - 1.0f) / 30.0f * Pi * 2;
		campos.x = sinf(w) * box_size/2;
		campos.y = 0;
		campos.z = cosf(w) * box_size/2;
		cam_viewpoint = Vec3::ZERO;
	}
	else if (t < 31.5f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = 1.0f;
		black_bars = true;
		scenetimer.setScale(0.1f, time);
	}
	else if (t < 32.0f) {
		campos = Vec3(0, 0, box_size/2 + box_cam_distance);
		fade = 1 - (t - 31.5f) / 0.5f;
		scenetimer.setScale(1.0f, time);
		black_bars = false;
	}
	else {
		black_bars = false;
		fade = 0.0f;
		campos = Vec3(0, 0, box_size/2 + box_cam_distance);
		scenetimer.setScale(1.0f, time);
		
		next_cam_effect = time + frand(90.0f, 150.0f);
		state = S_NORMAL;
		state_start = time;
	}
}

void CometenScene::processSlowmotion(float time) {
	float t = time - state_start;
	
	cam_viewpoint = Vec3::ZERO;
	
	if (t < 0.5f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = (t) / 0.5f;
		black_bars = false;
		scenetimer.setScale(1.0f, time);
	}
	else if (t < 1.0f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = 1.0f;
		black_bars = true;
		scenetimer.setScale(0.25f, time);
	}
	else if (t < 16.0f) {
		black_bars = true;
		if (t < 1.5f) {
			fade = 1 - (t - 1.0f) / 0.5f;
		}
		else if (t > 15.5f) {
			fade = (t - 15.5f) / 0.5f;
		}
		else {
			fade = 0.0f;
		}
		
		campos = Vec3(box_size/2 - 10, -box_height/2 + 10, box_size/2 - 10);
		scenetimer.setScale(0.25f, time);
	}
	else if (t < 31.0f) {
		black_bars = true;
		if (t < 16.5f) {
			fade = 1 - (t - 16.0f) / 0.5f;
		}
		else if (t > 30.5f) {
			fade = (t - 30.5f) / 0.5f;
		}
		else {
			fade = 0.0f;
		}
		
		campos = Vec3(-box_size/2 + 10, +box_height/2 - 10, -box_size/2 + 10);
		scenetimer.setScale(0.25f, time);
	}
	else if (t < 31.5f) {
		campos = Vec3(0,0,box_size/2 + box_cam_distance);
		fade = 1.0f;
		black_bars = true;
		scenetimer.setScale(0.25f, time);
	}
	else if (t < 32.0f) {
		fade = 1 - (t - 31.5f) / 0.5f;
		scenetimer.setScale(1.0f, time);
		black_bars = false;
		campos = Vec3(0, 0, box_size/2 + box_cam_distance);
	}
	else {
		black_bars = false;
		fade = 0.0f;
		campos = Vec3(0, 0, box_size/2 + box_cam_distance);
		scenetimer.setScale(1.0f, time);
		
		next_cam_effect = time + frand(90.0f, 150.0f);
		state = S_NORMAL;
		state_start = time;
	}
}

void CometenScene::processScene(float time) {
	Vec4 acolor, ecolor;
	
	processComets(time);
	processEffects(time);
	
	if (blitz) {
		processBlitz(time);
	}
	
	if (do_blitz && time > next_blitz_effect) {
		blitz = true;
		next_blitz_effect = time + frand(15.0f, 45.0f);
	}
	
	if (time > next_comet) {
		Comet * newc = NULL;
		
		switch (colors) {
		case C_RED:
			acolor = Vec4(1.0f, 0.5f, 0.25f, 1.0f);
			ecolor = Vec4(1.0f, 0.0f, 0.0f, 0.0f);
			break;
		case C_BLUE:
			acolor = Vec4(0.25f, 0.5f, 1.0f, 1.0f);
			ecolor = Vec4(0.0f, 0.0f, 1.0f, 0.0f);
			break;
		case C_GREEN:
			acolor = Vec4(0.25f, 1.0f, 0.65f, 1.0f);
			ecolor = Vec4(0.0f, 0.6f, 0.1f, 0.0f);
			break;
		case C_EXTREMBUNT:
			acolor = Vec4(frand(), frand(), frand(), 1.0f);
			ecolor = Vec4(frand(), frand(), frand(), 0.0f);
			break;
		case C_BUNT:
		default:
			switch (rand() % 4) {
			case 2:
				acolor = Vec4(1.0f, 0.5f, 0.25f, 1.0f);
				ecolor = Vec4(1.0f, 0.0f, 0.0f, 0.0f);
				break;
			case 1:
				acolor = Vec4(0.25f, 0.5f, 1.0f, 1.0f);
				ecolor = Vec4(0.0f, 0.0f, 1.0f, 0.0f);
				break;
			case 0:
				acolor = Vec4(0.25f, 1.0f, 0.65f, 1.0f);
				ecolor = Vec4(0.0f, 0.6f, 0.1f, 0.0f);
				break;
			case 3:
			default:
				acolor = Vec4(1.0f, 1.0f, 1.0f, 1.0f);
				ecolor = Vec4(0.0f, 0.0f, 0.5f, 0.0f);
				break;
			}
		}
		
		newc = new 
				PComet(this, time, frand(2.5f, 7.5f), rand(15,25), Vec3::ZERO, 
					   Vec3::randomUnit() * frand(75.0f, 100.0f), Vec3::ZERO, frand(0.75f, 1.5f),
					   acolor, ecolor);
	
		addComet(newc);
	
		next_comet = time + frand(create_intervall*0.75f, create_intervall*1.25f);
	}
}
	
void CometenScene::processBlitz(float time) {
	Effect * newe;
	list<Comet*>::iterator j;
	Comet * comet = NULL;
	
	for (j = comets.begin(); j != comets.end(); j++) {
		if (!(*j)->destroyed() && (*j)->getRadius() >= 0.75f) {
			comet = *j;
			break;
		}
	}
	
	if (comet == NULL) {
		blitz = false;
	}
	else {
		if (time - last_blitz > 0.3) {
			last_blitz = time;
			comet->destroy(time);
			
			newe = new Lightning(getTexture("particle.png"), 
			                     Vec3(0.0f, getBoxHeight() / 2, 0.0f), 
			                     comet->getPos(), 3.5f, 50, 3, time, 0.3f, 0.05f, 0.03f);
			addEffect(newe);
			newe = new TimeLimitDecal(getTexture("particle.png"), 
										Vec3(0.0f, getBoxHeight() / 2, 0.0f), false, 
										Vec3(0,-1,0), Vec2(12,12), 
										Vec2(25,25), Vec4(1.0f, 1.0f, 1.0f, 1.0f), time, 
										3.0f, 0.125f, 2.875f);
			addEffect(newe);
		}
	}			
}


	
void CometenScene::processComets(float time) {
	list<Comet*>::iterator j, k;
	list<Plane>::iterator plane, hit_plane;
	Comet *comet;
	Vec3 velocity;
	float deltat = time - last_comet_process;
	last_comet_process = time;
	
	for (j = comets.begin(); j != comets.end(); j++) {
		comet = *j;

		if (comet->done()) 
			continue;
		
		comet->process(time, deltat);
	}

	collision(time, deltat);
	
	for (j = comets.begin(); j != comets.end();) {
		comet = *j;
		
		if (comet == NULL || comet->done()) {
			delete comet;
			k = j;
			j++;
			comets.erase(k);
		}
		else {
			j++;
		}
	}
}

void CometenScene::processEffects(float time) {
	list<Effect*>::iterator i,k;
	Effect * effect;
	
	for (i = effects.begin(); i != effects.end();) {
		effect = *i;
		
		effect->process(time);
		
		if (effect->done()) { 
			delete effect;
			k = i;
			i++;
			effects.erase(k);
		}
		else { 
			i++;
		}
	}
}

void CometenScene::collision(float time, float deltat) {
	list<Comet*>::iterator j, k;
	list<Plane>::iterator plane, hit_plane;
	float t, hit_time;
	Comet *comet;
	
	for (j = comets.begin(); j != comets.end(); j++) {
		comet = *j;
		
		if (comet->destroyed()) 
			continue;
		
		//test planes
		hit_plane = planes.end();
		hit_time = deltat;

		for (plane = planes.begin(); plane != planes.end(); plane++) {
			// kill comets outside of the box
			if (comet->getPos().dot(plane->N) < plane->d) {
				comet->kill();
				break;
			}
			
			if (TestDynamicSpherePlane(comet->getPos(),
			                           comet->getRadius(),
			                           comet->getVel(), *plane, t) && 
			    t <= hit_time) {
					
				hit_plane = plane;
				hit_time = t;
			}
		} 
		
		if (comet->destroyed()) 
			continue;
		
		comet->setPos(comet->getPos() + comet->getVel() * hit_time);
		
		if (hit_plane != planes.end()) {
			cometHitPlane(time + hit_time, comet, *hit_plane);
		}
	}
}

void CometenScene::cometHitPlane(float time, Comet * comet, const Plane & plane) {
	float distance;
	const float e = 0.0001f;
	Vec3 hit_point;	
		
	distance = comet->getPos().dot(plane.N) - plane.d; 	
	
	hit_point = comet->getPos() - plane.N * (distance - 0.01f);
	
	// if comet intersects plane move along normal
	if (distance >= 0 && distance <= comet->getRadius() + e) {
		comet->setPos(comet->getPos() + (plane.N * (comet->getRadius() + e - distance)));
	}
	else if (distance < 0 && distance >= -comet->getRadius() - e) {
		comet->setPos(comet->getPos() + (plane.N * (-comet->getRadius() - e + distance)));
	}

	comet->hitPlane(time, plane, hit_point);
}
	
// ============================== add comet / effect ================================

void CometenScene::addComet(Comet* comet) {
	if (comet != NULL) 
		comets.push_front(comet);
}

void CometenScene::addEffect(Effect * effect) {
	if (effect != NULL) 
		effects.push_front(effect);
}

// =========================================== render ========================================

void CometenScene::render() {
	list<Comet*>::iterator i;
	list<Effect*>::iterator j;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(campos.x, campos.y, campos.z,
	          cam_viewpoint.x, cam_viewpoint.y, cam_viewpoint.z, 
	          0.0f, 1.0f, 0.0f);
	
	renderBackground();
	
	for (i = comets.begin(); i != comets.end(); i++) {
		(*i)->render();
	}
	
	for (j = effects.begin(); j != effects.end(); j++) {
		(*j)->render(campos);
	}
	
	glPopMatrix();
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.0, 640.0, 0.0, 480.0, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glDisable(GL_CULL_FACE);
	glDisable(GL_TEXTURE_2D);
	
	if (fade != 0.0f) {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glColor4f(0.0f, 0.0f, 0.0f, fade);
		glBegin(GL_QUADS);
		glVertex2f(0.0f, 0.0f);
		glVertex2f(0.0f, 480.0f);
		glVertex2f(640.0f, 480.0f);
		glVertex2f(640.0f, 0.0f);
		glEnd();

	}
	if (black_bars) {
		glDisable(GL_BLEND);
		glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
		float bar_height = 640.0f * 3 / 32;
		glBegin(GL_QUADS);
		glVertex2f(0.0f, 0.0f);
		glVertex2f(0.0f, bar_height);
		glVertex2f(640.0f, bar_height);
		glVertex2f(640.0f, 0.0f);
		glVertex2f(0.0f, 480.0f);
		glVertex2f(0.0f, 480.0f-bar_height);
		glVertex2f(640.0f, 480.0f-bar_height);
		glVertex2f(640.0f, 480.0f);
		glEnd();
	}
	
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void CometenScene::renderBackground() {
	float vertex_array[] = {
		// vorne
		box_size/2, -box_height / 2, box_size/2,
		box_size/2,  box_height / 2, box_size/2,
		-box_size/2,  box_height / 2, box_size/2,
		-box_size/2, -box_height / 2, box_size/2,	
		// hinten
		-box_size/2, -box_height / 2, -box_size/2,
		-box_size/2,  box_height / 2, -box_size/2,
		box_size/2,  box_height / 2, -box_size/2,
		box_size/2, -box_height / 2, -box_size/2,
		// links
		-box_size/2, -box_height / 2, box_size/2,
		-box_size/2,  box_height / 2, box_size/2,
		-box_size/2,  box_height / 2, -box_size/2,
		-box_size/2, -box_height / 2, -box_size/2,	
		// rechts
		box_size/2, -box_height / 2, -box_size/2,
		box_size/2,  box_height / 2, -box_size/2,	
		box_size/2,  box_height / 2, box_size/2,
		box_size/2, -box_height / 2, box_size/2,	
		// oben
		-box_size/2,  box_height / 2, -box_size/2,
		-box_size/2,  box_height / 2, box_size/2,	
		box_size/2,  box_height / 2, box_size/2,
		box_size/2,  box_height / 2, -box_size/2,	
		// unten
		-box_size/2, -box_height / 2, box_size/2,
		-box_size/2, -box_height / 2, -box_size/2,
		box_size/2, -box_height / 2, -box_size/2,
		box_size/2, -box_height / 2, box_size/2,
	};	
	
	static const float texcoords_array[] = {
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f,
	};
	
	glEnable(GL_TEXTURE_2D);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);		
	glTexCoordPointer(2, GL_FLOAT, 0, texcoords_array);
	glVertexPointer(3, GL_FLOAT, 0, vertex_array);
	glEnable(GL_CULL_FACE);
	glColor4f(0.7f, 0.7f, 0.7f, 1.0f);
	glBindTexture(GL_TEXTURE_2D, background);
	glDisable(GL_BLEND);	
	glDrawArrays(GL_QUADS, 0, 24);
	
	glEnable(GL_CULL_FACE);
	glTexCoordPointer(2, GL_FLOAT, 0, texcoords_array);
	glColor4f(1.0f, 1.0f, 1.0f, lightmap_alpha);
	glBindTexture(GL_TEXTURE_2D, lightmap);
	glEnable(GL_BLEND);	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDrawArrays(GL_QUADS, 0, 24);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

// ============================== load textures =====================================

GLuint CometenScene::getTexture(const string & file) {
	map<string, GLuint>::iterator k;
	
	k = textures.find(file);
	
	if (k != textures.end()) {
		return k->second;
	} else {
		return 0;
	}
}

void CometenScene::loadTexture(const string & file) {
	map<string, GLuint>::iterator k;
	
	k = textures.find(file);
	
	if (k != textures.end()) {
		freeTexture(file);	
	}
	
	textures[file] = load_texture(file);
}

void CometenScene::freeTexture(const string& file) {
	map<string, GLuint>::iterator k;
	
	k = textures.find(file);
	
	if (k != textures.end()) {
		if (k->second != 0) 
			glDeleteTextures(1, &(k->second));
		textures.erase(k);
	}
}

void CometenScene::freeTextures() {
	map<string, GLuint>::iterator k;
	
	for (k = textures.begin(); k != textures.end(); k++) {
		if (k->second != 0) 
			glDeleteTextures(1, &(k->second));
	}
	
	textures.clear();
}
	

	
	
	
	
	
	
	
	
