/***************************************************************************
 *   Copyright (C) 2005 by Peter Müller   *
 *   pmueller@cs.tu-berlin.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "decal.h"

Decal::Decal(GLuint _texture, const Vec3 & _pos, bool _billboard,
             const Vec3 & _normal, const Vec2 & _size, const Vec4 & _color,
             float time, float _fadein_time, float _fadeout_time)
:
Effect(_pos),
texture(_texture),
billboard(_billboard),
normal(_normal),
size(_size),
color(_color),
start_time(time),
fadein_time(_fadein_time),
fadeout_time(_fadeout_time),
stopped(false)
{
	transparency = 0.0f;
}


Decal::~Decal()
{
}

void Decal::process(float time) {
	float s;
	
	last_process_time = time;
	
	if (stopped && time > end_time) {
		transparency = 0.0f;
	} 
	else if (stopped && time > end_time - fadeout_time) {
		// fadeout t 1->0
		transparency = 1 - (time - end_time + fadeout_time) / fadeout_time;
	}
	else if (time < start_time + fadein_time) {
		// fadein t 0->1
		transparency = (time - start_time) / fadein_time;
	}
	else {
		transparency = 1;
	}
	

}

void Decal::render(const Vec3 & campos) {
	Vec3 dir, up, right;
	float dotDY;
		
	if (billboard) {
		dir = campos - pos;
		dir.normalize();
	} else {
		dir = normal;
	}
	
	dotDY = dir.dot(Vec3::Y);
	
	if (dotDY == 1.0f || dotDY == -1.0f) {
		up = -Vec3::Z;
	}
	else {
		up = Vec3::Y - (dir * dotDY);
		up.normalize();
	}
	
	right = dir.cross(up);
	
	right *= 0.5f * size.x;
	up *= 0.5f * size.y;
	
	glPushMatrix();

	glTranslatef(pos.x, pos.y, pos.z);
	
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE );
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);
	
	glColor4f(color.x, color.y, color.z, color.w * transparency);
	
	glBegin(GL_QUADS);
	
	glTexCoord2f(0.0f, 1.0f);	
	glVertex3f((right.x + up.x), (right.y + up.y), (right.z + up.z));
	glTexCoord2f(1.0f, 1.0f);
	glVertex3f((-right.x + up.x), (-right.y + up.y), (-right.z + up.z));
	glTexCoord2f(1.0f, 0.0f);
	glVertex3f((-right.x - up.x), (-right.y - up.y), (-right.z - up.z));
	glTexCoord2f(0.0f, 0.0f);
	glVertex3f((right.x - up.x), (right.y - up.y), (right.z - up.z));
	
	glEnd();
	
	glPopMatrix();	
}

void Decal::stop() {
	if (!stopped) {
		stopped = true;
		end_time = last_process_time + fadeout_time;
	}
}

bool Decal::done() {
	return stopped && last_process_time > end_time;
}

TimeLimitDecal::TimeLimitDecal(GLuint _texture, const Vec3 & _pos, bool _billboard, const Vec3 & _normal,
                         const Vec2 & _start_size, const Vec2 & _end_size, const Vec4 & _color,
                         float time, float _life_time, float _fadein_time, float _fadeout_time)
: 
Decal(_texture, _pos, _billboard, _normal, _start_size, 
      _color, time, _fadein_time, _fadeout_time),
start_size(_start_size),
end_size(_end_size)
{
	end_time = time + _life_time;
	stopped = true;
}

TimeLimitDecal::~TimeLimitDecal() {
}

void TimeLimitDecal::process(float time, float deltat) {
	float s;
	
	last_process_time = time;
	
	if (stopped && time > end_time) {
		transparency = 0.0f;
	} 
	else if (stopped && time > end_time - fadeout_time) {
		// fadeout t 1->0
		transparency = 1 - (time - end_time + fadeout_time) / fadeout_time;
	}
	else if (time < start_time + fadein_time) {
		// fadein t 0->1
		transparency = (time - start_time) / fadein_time;
	}
	else {
		transparency = 1;
	}
	
	s = (time - start_time) / (end_time - start_time);
	size = start_size * (1-s) + end_size * s;
	
}

